import json
import re
from io import BytesIO
from urllib.parse import urlsplit

from django.http import HttpRequest
from django.http.request import QueryDict
from django.utils.encoding import force_text

from rest_framework.exceptions import ValidationError

from drf_batch_requests.exceptions import RequestAttributeError
from drf_batch_requests.serializers import BatchRequestSerializer
from drf_batch_requests.utils import get_attribute


class BatchRequest(HttpRequest):

    def __init__(self, request, request_data):
        super(BatchRequest, self).__init__()
        self.name = request_data.get('name')
        self.omit_response_on_success = request_data.get('omit_response_on_success', False)

        self._stream = BytesIO(request_data['_body'].encode('utf-8'))
        self._read_started = False

        self.method = request_data['method']

        split_url = urlsplit(request_data['relative_url'])
        self.path_info = self.path = split_url.path

        self.GET = QueryDict(split_url.query)
        self.META = request.META
        self.COOKIES = request.COOKIES


class BatchRequestsFactory(object):
    response_variable_regex = re.compile(r'({result=(?P<name>[\w\d_]+):\$\.(?P<value>[\w\d_.*]+)})')

    def __init__(self, request):
        self.request = request
        self.request_serializer = BatchRequestSerializer(data=request.data)
        self.request_serializer.is_valid(raise_exception=True)
        self.update_soft_dependencies()

        self.named_responses = {}

    def update_soft_dependencies(self):
        for request_data in self.request_serializer.validated_data['batch']:
            parents = request_data.get('depends_on', [])

            for part in request_data.values():
                params = re.findall(
                    self.response_variable_regex, force_text(part)
                )

                parents.extend(map(lambda param: param[1], params or []))

            request_data['depends_on'] = set(parents)

    def _prepare_formdata_body(self, data, files=None):
        if not data and not files:
            return ''

        match = re.search(r'boundary=(?P<boundary>.+)', self.request.content_type)
        assert match
        boundary = match.groupdict()['boundary']
        body = ''
        for key, value in data.items():
            value = value if isinstance(value, str) else json.dumps(value)
            body += '--{}\r\nContent-Disposition: form-data; name="{}"\r\n\r\n{}\r\n'.format(boundary, key, value)

        if files:
            for key, attachment in files.items():
                attachment.seek(0)
                attachment_body_part = '--{0}\r\nContent-Disposition: form-data; name="{1}"; filename="{2}"\r\n' \
                                       'Content-Type: {3}\r\n' \
                                       'Content-Transfer-Encoding: binary\r\n\r\n{4}\r\n'
                body += attachment_body_part.format(
                    boundary, key, attachment.name, attachment.content_type, attachment.read()
                )

        body += '--{}--\r\n'.format(boundary)
        return body

    def _prepare_urlencoded_body(self, data):
        raise NotImplementedError

    def _prepare_json_body(self, data):
        return json.dumps(data)

    def _process_attr(self, attr):
        params = re.findall(
            self.response_variable_regex, attr
        )
        if not params:
            return attr

        for url_param in params:
            if url_param[1] not in self.named_responses:
                raise ValidationError('Named request {} is missing'.format(url_param[1]))

            result = get_attribute(
                self.named_responses[url_param[1]]['_data'],
                url_param[2].split('.')
            )

            if result is None:
                raise RequestAttributeError('Empty result for {}'.format(url_param[2]))

            if isinstance(result, list):
                result = ','.join(map(str, result))

            if attr == url_param[0]:
                attr = result
            else:
                attr = attr.replace(url_param[0], str(result))

        return attr

    def updated_obj(self, obj):
        """
        For now, i'll update only dict values. Later it can be used for keys/single values/etc
        :param obj: dict
        :return: dict
        """
        if isinstance(obj, dict):
            for key, value in obj.items():
                obj[key] = self.updated_obj(value)
        elif isinstance(obj, str):
            return self._process_attr(obj)

        return obj

    def get_requests_data(self):
        return self.request_serializer.validated_data['batch']

    def generate_request(self, request_data):
        request_data['data'] = self.updated_obj(request_data['data'])
        request_data['relative_url'] = self._process_attr(request_data['relative_url'])

        if self.request.content_type.startswith('multipart/form-data'):
            request_data['_body'] = self._prepare_formdata_body(request_data['data'],
                                                                files=request_data.get('files', {}))
        elif self.request.content_type.startswith('application/x-www-form-urlencoded'):
            request_data['_body'] = self._prepare_urlencoded_body(request_data['data'])
        elif self.request.content_type.startswith('application/json'):
            request_data['_body'] = self._prepare_json_body(request_data['data'])
        else:
            raise ValidationError('Unsupported content type')

        return BatchRequest(self.request, request_data)



from django.core.handlers.base import BaseHandler

from rest_framework.status import is_success

from drf_batch_requests.backends.base import RequestsConsumeBaseBackend


class SyncRequestsConsumeBackend(RequestsConsumeBaseBackend):
    def __init__(self):
        self.responses = {}

    # todo: from this point i think we can consume requests pack
    def consume_request(self, request, start_callback=None, success_callback=None, fail_callback=None):
        start_callback() if start_callback else None

        handler = BaseHandler()
        handler.load_middleware()

        response = handler.get_response(request)

        if is_success(response.status_code):
            success_callback() if success_callback else None
        else:
            fail_callback() if fail_callback else None

        self.responses[request] = response

        return True



import json

from django.core.files import File

from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from drf_batch_requests.utils import generate_random_id


class SingleRequestSerializer(serializers.Serializer):
    name = serializers.CharField(required=False)
    depends_on = serializers.JSONField(required=False)
    method = serializers.CharField()
    relative_url = serializers.CharField()
    body = serializers.JSONField(required=False, default={})
    # attached files formats: ["a.jpg", "b.png"] - will be attached as it is, {"file": "a.jpg"} - attach as specific key
    attached_files = serializers.JSONField(required=False)
    data = serializers.SerializerMethodField()
    files = serializers.SerializerMethodField()

    def validate_relative_url(self, value):
        if not value.startswith('/'):
            raise self.fail('Url should start with /')

        return value

    def validate_body(self, value):
        if isinstance(value, dict):
            return value

        try:
            json.loads(value)
        except (TypeError, ValueError):
            self.fail('invalid')

        return value

    def validate(self, attrs):
        if 'name' not in attrs:
            attrs['name'] = generate_random_id()

        if 'depends_on' in attrs:
            value = attrs['depends_on']
            if not isinstance(value, (str, list)):
                raise ValidationError({'depends_on': 'Incorrect value provided'})

            if isinstance(value, str):
                attrs['depends_on'] = [value]

        return attrs

    def get_data(self, data):
        body = data['body']
        if isinstance(body, dict):
            return body

        return json.loads(body)

    def get_files(self, attrs):
        if 'attached_files' not in attrs:
            return []

        attached_files = attrs['attached_files']
        if isinstance(attached_files, dict):
            return {
                key: self.context['parent'].get_files()[attrs['attached_files'][key]] for key in attrs['attached_files']
            }
        elif isinstance(attached_files, list):
            return {
                key: self.context['parent'].get_files()[key] for key in attrs['attached_files']
            }
        else:
            raise ValidationError('Incorrect format.')


class BatchRequestSerializer(serializers.Serializer):
    batch = serializers.JSONField()
    files = serializers.SerializerMethodField()

    def get_files(self, attrs=None):
        return {fn: f for fn, f in self.initial_data.items() if isinstance(f, File)}

    def validate_batch(self, value):
        if not isinstance(value, list):
            raise ValidationError('List of requests should be provided to do batch')

        r_serializers = list(map(lambda d: SingleRequestSerializer(data=d, context={'parent': self}), value))

        errors = []
        for serializer in r_serializers:
            serializer.is_valid()
            errors.append(serializer.errors)
        if any(errors):
            raise ValidationError(errors)

        return [s.data for s in r_serializers]

    def validate(self, attrs):
        attrs = super(BatchRequestSerializer, self).validate(attrs)

        files_in_use = []
        for batch in attrs['batch']:
            if 'attached_files' not in batch:
                continue

            attached_files = batch['attached_files']
            if isinstance(attached_files, dict):
                files_in_use.extend(attached_files.values())
            elif isinstance(attached_files, list):
                files_in_use.extend(attached_files)
            else:
                raise ValidationError({'attached_files': 'Invalid format.'})

        missing_files = set(files_in_use) - set(self.get_files().keys())
        if missing_files:
            raise ValidationError('Some of files are not provided: {}'.format(', '.join(missing_files)))

        return attrs



import json
from importlib import import_module

from django.db import transaction

from rest_framework.response import Response
from rest_framework.status import is_success
from rest_framework.views import APIView

from drf_batch_requests import settings as app_settings
from drf_batch_requests.exceptions import RequestAttributeError
from drf_batch_requests.graph import RequestGraph
from drf_batch_requests.request import BatchRequestsFactory
from drf_batch_requests.utils import generate_node_callback

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


class BatchView(APIView):
    permission_classes = []

    def get_requests_consumer_class(self):
        mod, inst = app_settings.REQUESTS_CONSUMER_BACKEND.rsplit('.', 1)
        mod = import_module(mod)
        return getattr(mod, inst)

    def get_requests_consumer(self):
        return self.get_requests_consumer_class()()

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        requests = {}
        responses = {}

        requests_factory = BatchRequestsFactory(request)
        requests_data = requests_factory.get_requests_data()
        ordered_names = list(map(lambda r: r['name'], requests_data))
        requests_graph = RequestGraph(requests_data)

        backend = self.get_requests_consumer()

        while True:
            available_nodes = list(requests_graph.get_current_available_nodes())

            for node in available_nodes:
                try:
                    current_request = requests_factory.generate_request(node.request)
                except RequestAttributeError:
                    # todo: set fail reason
                    node.fail()

                start_callback = generate_node_callback(node, 'start')
                success_callback = generate_node_callback(node, 'success')
                fail_callback = generate_node_callback(node, 'fail')
                if backend.consume_request(current_request, start_callback=start_callback,
                                           success_callback=success_callback, fail_callback=fail_callback):
                    requests[node.name] = current_request

            is_completed = requests_graph.is_completed()

            for current_request, response in backend.responses.items():
                if current_request.name in responses:
                    continue

                result = {
                    'name': current_request.name,
                    'code': response.status_code,
                    'headers': [
                        {'name': key, 'value': value}
                        for key, value in response._headers.values()
                    ],
                    'body': response.content.decode('utf-8'),
                }

                if is_success(response.status_code):
                    try:
                        result['_data'] = json.loads(result['body'])
                    except JSONDecodeError:
                        pass

                if not is_success(response.status_code) or \
                   is_success(response.status_code) and not current_request.omit_response_on_success:
                    result['return_body'] = True

                if current_request.name:
                    requests_factory.named_responses[current_request.name] = result

                responses[current_request.name] = result

            if is_completed:
                break

        ordered_responses = [responses.get(name, {'name': name, 'code': 418}) for name in ordered_names]
        return self.finalize_response(request, Response(ordered_responses))



from django.conf import settings

# Consumer backend
REQUESTS_CONSUMER_BACKEND = getattr(
    settings, "DRF_BATCH_REQUESTS_CONSUMER_BACKEND", 'drf_batch_requests.backends.sync.SyncRequestsConsumeBackend'
)



import os

from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'readme.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='drf-batch-requests',
    version='0.9.1',
    packages=['drf_batch_requests', ],
    include_package_data=True,
    license='MIT License',
    description='Facebook-like implementation of batch requests',
    long_description=README,
    keywords='django django-rest-framework drf batch',
    url='https://github.com/roman-karpovich/drf-batch-requests',
    author='Roman Karpovich',
    author_email='fpm.th13f@gmail.com',
    install_requires=[
        'django>=2.2',
        'djangorestframework',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    zip_safe=True
)



try:
    from django.conf.urls import url
except ImportError:
    # django 2.0
    from django.urls import re_path as url

from drf_batch_requests import views

app_name = 'drt_batch_requests'

urlpatterns = [
    url('^', views.BatchView.as_view())
]



import random
import string


def get_attribute(instance, attrs):
    for attr in attrs:
        if instance is None:
            return None

        if attr == '*':
            # todo: maybe there should be some kind of filtering?
            continue

        if isinstance(instance, list):
            instance = list(map(lambda i: i[attr], instance))
        else:
            instance = instance[attr]
    return instance


def generate_random_id(size=10, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))


def generate_node_callback(node, status):
    def callback():
        if status == 'start':
            node.start()
        elif status == 'success':
            node.complete()
        elif status == 'fail':
            node.fail()
        else:
            raise NotImplementedError

    return callback



class RequestGraphNode(object):
    STATUS_FAILED = -1
    STATUS_FAILED_PARENT = -2
    STATUS_NOT_STARTED = 0
    STATUS_IN_PROGRESS = 1
    STATUS_COMPLETED = 2

    def __init__(self, request=None):
        self.request = request
        self.name = self.request.get('name')
        self.parents = set()
        self.children_set = set()
        self.status = self.STATUS_NOT_STARTED

    def start(self):
        self.status = RequestGraphNode.STATUS_IN_PROGRESS

    def complete(self):
        self.status = RequestGraphNode.STATUS_COMPLETED

    def fail(self, own_fail=True):
        self.status = RequestGraphNode.STATUS_FAILED if own_fail else RequestGraphNode.STATUS_FAILED_PARENT
        for child_node in filter(lambda n: n.status == RequestGraphNode.STATUS_NOT_STARTED, self.children_set):
            child_node.fail(own_fail=False)

    @property
    def can_be_performed(self):
        if not self.status == RequestGraphNode.STATUS_NOT_STARTED:
            return False

        return all(map(lambda parent: parent.status == RequestGraphNode.STATUS_COMPLETED, self.parents))

    def __str__(self):
        return self.name or super(RequestGraphNode, self).__str__()


class RequestGraph(object):
    def __init__(self, requests):
        self.nodes = [RequestGraphNode(request) for request in requests]
        self._named_requests = {
            node.request['name']: node
            for node in filter(lambda n: n.request.get('name'), self.nodes)
        }

        for node in self.nodes:
            parents = node.request.get('depends_on', [])

            for parent_name in parents:
                parent = self._named_requests.get(parent_name)
                if not parent:
                    raise Exception('Wrong parent {} in node.'.format(parent_name))

                node.parents.add(parent)
                parent.children_set.add(node)

    def get_node_order(self, node):
        return self.nodes.index(node)

    def get_not_failed_nodes(self):
        return filter(
            lambda node: node.status not in [
                RequestGraphNode.STATUS_FAILED,
                RequestGraphNode.STATUS_FAILED_PARENT
            ],
            self.nodes
        )

    def get_current_available_nodes(self):
        return filter(lambda node: node.can_be_performed, self.get_not_failed_nodes())

    def is_completed(self):
        return all(map(
            lambda node: node.status in [
                RequestGraphNode.STATUS_FAILED,
                RequestGraphNode.STATUS_FAILED_PARENT,
                RequestGraphNode.STATUS_COMPLETED
            ],
            self.nodes
        ))
